@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2016-2020 Stephane Caron <stephane.caron@normalesup.org>
+# Copyright (C) 2016-2021 Stéphane Caron <stephane.caron@normalesup.org>
 #
 # This file is part of qpsolvers.
 #
@@ -18,160 +18,127 @@
 # You should have received a copy of the GNU Lesser General Public License
 # along with qpsolvers. If not, see <http://www.gnu.org/licenses/>.
 
-from numpy import concatenate, eye, ndarray
+"""Quadratic programming solvers in Python with a unified API"""
+
+from typing import Any, Dict, Optional
+
+from numpy import concatenate, dot, eye, hstack, ones, ndarray, vstack, zeros
+
+__version__ = "1.6"
 
 available_solvers = []
 dense_solvers = []
 sparse_solvers = []
+__solve_function__: Dict[str, Any] = {}
+
 
 # CVXOPT
 # ======
 
 try:
-    from .cvxopt_ import cvxopt_set_verbosity
     from .cvxopt_ import cvxopt_solve_qp
 
-    available_solvers.append('cvxopt')
-    dense_solvers.append('cvxopt')
+    __solve_function__["cvxopt"] = cvxopt_solve_qp
+    available_solvers.append("cvxopt")
+    dense_solvers.append("cvxopt")
 except ImportError:
-    def cvxopt_set_verbosity(*args, **kwargs):
-        pass
-
-
-    def cvxopt_solve_qp(*args, **kwargs):
-        raise ImportError("CVXOPT not found")
+    pass
 
 # CVXPY
 # =====
 
 try:
-    from .cvxpy_ import cvxpy_set_verbosity
     from .cvxpy_ import cvxpy_solve_qp
 
-    available_solvers.append('cvxpy')
-    sparse_solvers.append('cvxpy')
+    __solve_function__["cvxpy"] = cvxpy_solve_qp
+    available_solvers.append("cvxpy")
+    sparse_solvers.append("cvxpy")
 except ImportError:
-    def cvxpy_set_verbosity(*args, **kwargs):
-        pass
-
-
-    def cvxpy_solve_qp(*args, **kwargs):
-        raise ImportError("CVXPY not found")
+    pass
 
 # ECOS
 # ====
 
 try:
-    from .ecos_ import ecos_set_verbosity
     from .ecos_ import ecos_solve_qp
 
-    available_solvers.append('ecos')
-    dense_solvers.append('ecos')  # considered dense as it calls cholesky(P)
+    __solve_function__["ecos"] = ecos_solve_qp
+    available_solvers.append("ecos")
+    dense_solvers.append("ecos")  # considered dense as it calls cholesky(P)
 except ImportError:
-    def ecos_set_verbosity(*args, **kwargs):
-        pass
-
-
-    def ecos_solve_qp(*args, **kwargs):
-        raise ImportError("ECOS not found")
+    pass
 
 # Gurobi
 # ======
 
 try:
-    from .gurobi_ import gurobi_set_verbosity
     from .gurobi_ import gurobi_solve_qp
 
-    available_solvers.append('gurobi')
-    sparse_solvers.append('gurobi')
+    __solve_function__["gurobi"] = gurobi_solve_qp
+    available_solvers.append("gurobi")
+    sparse_solvers.append("gurobi")
 except ImportError:
-    def gurobi_set_verbosity(*args, **kwargs):
-        pass
+    pass
 
-
-    def gurobi_solve_qp(*args, **kwargs):
-        raise ImportError("Gurobi not found")
-
-# Mosek
+# MOSEK
 # =====
 
 try:
-    from .mosek_ import mosek_set_verbosity
     from .mosek_ import mosek_solve_qp
 
-    available_solvers.append('mosek')
-    sparse_solvers.append('mosek')
+    __solve_function__["mosek"] = mosek_solve_qp
+    available_solvers.append("mosek")
+    sparse_solvers.append("mosek")
 except ImportError:
-    def mosek_set_verbosity(*args, **kwargs):
-        pass
-
-
-    def mosek_solve_qp(*args, **kwargs):
-        raise ImportError("mosek not found")
+    pass
 
 # OSQP
 # ====
 
 try:
-    from .osqp_ import osqp_set_verbosity
     from .osqp_ import osqp_solve_qp
 
-    available_solvers.append('osqp')
-    sparse_solvers.append('osqp')
+    __solve_function__["osqp"] = osqp_solve_qp
+    available_solvers.append("osqp")
+    sparse_solvers.append("osqp")
 except ImportError:
-    def osqp_set_verbosity(*args, **kwargs):
-        pass
-
-
-    def osqp_solve_qp(*args, **kwargs):
-        raise ImportError("osqp not found")
+    pass
 
 # qpOASES
 # =======
 
 try:
-    from .qpoases_ import qpoases_set_verbosity
     from .qpoases_ import qpoases_solve_qp
 
-    available_solvers.append('qpoases')
-    dense_solvers.append('qpoases')
+    __solve_function__["qpoases"] = qpoases_solve_qp
+    available_solvers.append("qpoases")
+    dense_solvers.append("qpoases")
 except ImportError:
-    def qpoases_set_verbosity(*args, **kwargs):
-        pass
-
-
-    def qpoases_solve_qp(*args, **kwargs):
-        raise ImportError("qpOASES not found")
+    pass
 
 # quadprog
 # ========
 
 try:
-    from .quadprog_ import quadprog_set_verbosity
     from .quadprog_ import quadprog_solve_qp
 
-    available_solvers.append('quadprog')
-    dense_solvers.append('quadprog')
+    __solve_function__["quadprog"] = quadprog_solve_qp
+    available_solvers.append("quadprog")
+    dense_solvers.append("quadprog")
 except ImportError:
-    def quadprog_set_verbosity(*args, **kwargs):
-        pass
+    pass
 
 
-    def quadprog_solve_qp(*args, **kwargs):
-        raise ImportError("quadprog not found")
+class SolverNotFound(Exception):
+    pass
 
 
-def check_problem(P, q, G, h, A, b, lb, ub):
+def check_problem_constraints(G, h, A, b) -> None:
     """
-    Check that problem matrices and vectors are correctly defined.
+    Check that problem constraint matrices and vectors are correctly defined.
 
     Parameters
     ----------
-    P : numpy.array, scipy.sparse.csc_matrix or cvxopt.spmatrix
-        Symmetric quadratic-cost matrix (most solvers require it to be definite
-        as well).
-    q : numpy.array
-        Quadratic-cost vector.
     G : numpy.array, scipy.sparse.csc_matrix or cvxopt.spmatrix
         Linear inequality matrix.
     h : numpy.array
@@ -180,33 +147,50 @@ def check_problem(P, q, G, h, A, b, lb, ub):
         Linear equality matrix.
     b : numpy.array
         Linear equality vector.
-    lb: numpy.array, scipy.sparse.csc_matrix or cvxopt.spmatrix
-        Lower bound constraint vector.
-    ub: numpy.array, scipy.sparse.csc_matrix or cvxopt.spmatrix
-        Upper bound constraint vector.
+
+    Raises
+    ------
+    ValueError
+        If the constraints are not properly defined.
     """
     if G is None and h is not None:
         raise ValueError("incomplete inequality constraint (missing h)")
-    elif G is not None and h is None:
+    if G is not None and h is None:
         raise ValueError("incomplete inequality constraint (missing G)")
     if A is None and b is not None:
         raise ValueError("incomplete equality constraint (missing b)")
-    elif A is not None and b is None:
+    if A is not None and b is None:
         raise ValueError("incomplete equality constraint (missing A)")
 
 
-def solve_qp(P, q, G=None, h=None, A=None, b=None, lb=None, ub=None,
-             solver='quadprog', initvals=None, sym_proj=False, verbose=False):
+def solve_qp(
+    P,
+    q,
+    G=None,
+    h=None,
+    A=None,
+    b=None,
+    lb=None,
+    ub=None,
+    solver="quadprog",
+    initvals=None,
+    sym_proj=False,
+    verbose=False,
+    **kwargs,
+) -> Optional[ndarray]:
     """
     Solve a Quadratic Program defined as:
 
-        minimize
-            (1/2) * x.T * P * x + q.T * x
+    .. math::
 
-        subject to
-            lb <= x <= ub
-            G * x <= h
-            A * x == b
+        \\begin{split}\\begin{array}{ll}
+            \\mbox{minimize} &
+                \\frac{1}{2} x^T P x + q^T x \\\\
+            \\mbox{subject to}
+                & G x \\leq h                \\\\
+                & A x = b                    \\\\
+                & lb \\leq x \\leq ub
+        \\end{array}\\end{split}
 
     using one of the available QP solvers.
 
@@ -217,17 +201,17 @@ def solve_qp(P, q, G=None, h=None, A=None, b=None, lb=None, ub=None,
         as well).
     q : numpy.array
         Quadratic-cost vector.
-    G : numpy.array, scipy.sparse.csc_matrix or cvxopt.spmatrix
+    G : numpy.array, scipy.sparse.csc_matrix or cvxopt.spmatrix, optional
         Linear inequality matrix.
-    h : numpy.array
+    h : numpy.array, optional
         Linear inequality vector.
-    A : numpy.array, scipy.sparse.csc_matrix or cvxopt.spmatrix
+    A : numpy.array, scipy.sparse.csc_matrix or cvxopt.spmatrix, optional
         Linear equality matrix.
-    b : numpy.array
+    b : numpy.array, optional
         Linear equality vector.
-    lb: numpy.array, scipy.sparse.csc_matrix or cvxopt.spmatrix
+    lb: numpy.array, scipy.sparse.csc_matrix or cvxopt.spmatrix, optional
         Lower bound constraint vector.
-    ub: numpy.array, scipy.sparse.csc_matrix or cvxopt.spmatrix
+    ub: numpy.array, scipy.sparse.csc_matrix or cvxopt.spmatrix, optional
         Upper bound constraint vector.
     solver : string, optional
         Name of the QP solver, to choose in ``qpsolvers.available_solvers``.
@@ -243,6 +227,17 @@ def solve_qp(P, q, G=None, h=None, A=None, b=None, lb=None, ub=None,
     x : array or None
         Optimal solution if found, None otherwise.
 
+    Raises
+    ------
+    ValueError
+        If the problem is not correctly defined.
+
+    Note
+    ----
+    Extra keyword arguments given to this function are forwarded to the
+    underlying solvers. For example, OSQP has a setting `eps_abs` which we can
+    provide by ``solve_qp(P, q, G, h, solver='osqp', eps_abs=1e-4)``.
+
     Notes
     -----
     In quadratic programming, the matrix `P` should be symmetric. Many solvers
@@ -251,12 +246,12 @@ def solve_qp(P, q, G=None, h=None, A=None, b=None, lb=None, ub=None,
     project `P` on its symmetric part, at the cost of some computation time.
     """
     if sym_proj:
-        P = .5 * (P + P.transpose())
-    if type(A) is ndarray and A.ndim == 1:
+        P = 0.5 * (P + P.transpose())
+    if isinstance(A, ndarray) and A.ndim == 1:
         A = A.reshape((1, A.shape[0]))
-    if type(G) is ndarray and G.ndim == 1:
+    if isinstance(G, ndarray) and G.ndim == 1:
         G = G.reshape((1, G.shape[0]))
-    check_problem(P, q, G, h, A, b, lb, ub)
+    check_problem_constraints(G, h, A, b)
     if lb is not None:
         if G is None:
             G = -eye(len(q))
@@ -271,43 +266,39 @@ def solve_qp(P, q, G=None, h=None, A=None, b=None, lb=None, ub=None,
         else:  # G is not None and h is not None
             G = concatenate((G, eye(len(q))), 0)
             h = concatenate((h, ub))
-    if solver == 'cvxopt':
-        cvxopt_set_verbosity(verbose)
-        return cvxopt_solve_qp(P, q, G, h, A, b, initvals=initvals)
-    elif solver == 'cvxpy':
-        cvxpy_set_verbosity(verbose)
-        return cvxpy_solve_qp(P, q, G, h, A, b, initvals=initvals)
-    elif solver == 'ecos':
-        ecos_set_verbosity(verbose)
-        return ecos_solve_qp(P, q, G, h, A, b, initvals=initvals)
-    elif solver == 'gurobi':
-        gurobi_set_verbosity(verbose)
-        return gurobi_solve_qp(P, q, G, h, A, b, initvals=initvals)
-    elif solver == 'mosek':
-        mosek_set_verbosity(verbose)
-        return mosek_solve_qp(P, q, G, h, A, b, initvals=initvals)
-    elif solver == 'osqp':
-        osqp_set_verbosity(verbose)
-        return osqp_solve_qp(P, q, G, h, A, b, initvals=initvals)
-    elif solver == 'qpoases':
-        qpoases_set_verbosity(verbose)
-        return qpoases_solve_qp(P, q, G, h, A, b, initvals=initvals)
-    elif solver == 'quadprog':
-        quadprog_set_verbosity(verbose)
-        return quadprog_solve_qp(P, q, G, h, A, b, initvals=initvals)
-    raise Exception("solver '%s' not recognized" % solver)
+    args = P, q, G, h, A, b
+    kwargs["initvals"] = initvals
+    kwargs["verbose"] = verbose
+    try:
+        return __solve_function__[solver](*args, **kwargs)
+    except KeyError:
+        raise SolverNotFound(f"solver '{solver}' is not available")
 
 
-def solve_safer_qp(P, q, G, h, sw, reg=1e-8, solver='mosek', initvals=None,
-                   sym_proj=False):
+def solve_safer_qp(
+    P,
+    q,
+    G,
+    h,
+    sw: float,
+    reg: float = 1e-8,
+    solver="mosek",
+    initvals=None,
+    sym_proj=False,
+) -> Optional[ndarray]:
     """
     Solve the Quadratic Program defined as:
 
-        minimize
-            (1/2) * x.T * P * x + q.T * x + (1/2) reg |s|^2 - sw 1^T s
+    .. math::
 
-        subject to
-            G * x <= h
+        \\begin{split}\\begin{array}{ll}
+            \\mbox{minimize} &
+                \\frac{1}{2} x^T P x + q^T x +
+                \\frac{1}{2} \\mathit{reg} \\|s\\|^2 - \\mathit{sw} \\1^T s
+                \\\\
+            \\mbox{subject to}
+                & G x \\leq h
+        \\end{array}\\end{split}
 
     Slack variables `s` are increased by an additional term in the cost
     function, so that the solution of this "safer" QP is further inside the
@@ -323,11 +314,11 @@ def solve_safer_qp(P, q, G, h, sw, reg=1e-8, solver='mosek', initvals=None,
         Linear inequality matrix.
     h : numpy.array
         Linear inequality vector.
-    sw : scalar
+    sw : float
         Weight of the linear cost on slack variables. Higher values bring the
         solution further inside the constraint region but override the
         minimization of the original objective.
-    reg : scalar
+    reg : float, optional
         Regularization term :math:`(1/2) \\epsilon` in the cost function. Set
         this parameter as small as possible (e.g. 1e-8), and increase it in
         case of numerical instability.
@@ -356,7 +347,7 @@ def solve_safer_qp(P, q, G, h, sw, reg=1e-8, solver='mosek', initvals=None,
     Transactions on Robotics, 2009).
     """
     assert solver in dense_solvers, "only available for dense solvers, for now"
-    from numpy import eye, hstack, ones, vstack, zeros
+
     n, m = P.shape[0], G.shape[0]
     E, Z = eye(m), zeros((m, n))
     P2 = vstack([hstack([P, Z.T]), hstack([Z, reg * eye(m)])])
@@ -366,21 +357,122 @@ def solve_safer_qp(P, q, G, h, sw, reg=1e-8, solver='mosek', initvals=None,
     A2 = hstack([G, -E])
     b2 = h
     x = solve_qp(
-        P2, q2, G2, h2, A2, b2, solver=solver, initvals=initvals,
-        sym_proj=sym_proj)
+        P2, q2, G2, h2, A2, b2, solver=solver, initvals=initvals, sym_proj=sym_proj
+    )
+    if x is None:
+        return None
     return x[:n]
 
 
+def solve_ls(
+    R,
+    s,
+    G=None,
+    h=None,
+    A=None,
+    b=None,
+    lb=None,
+    ub=None,
+    W=None,
+    solver="quadprog",
+    initvals=None,
+    sym_proj=False,
+    verbose=False,
+    **kwargs,
+) -> Optional[ndarray]:
+    """
+    Solve a constrained weighted linear Least Squares problem defined as:
+
+    .. math::
+
+        \\begin{split}\\begin{array}{ll}
+            \\mbox{minimize} &
+                \\frac12 \\| R x - s \\|^2_W
+                = \\frac12 (R x - s)^T W (R x - s) \\\\
+            \\mbox{subject to}
+                & G x \\leq h                \\\\
+                & A x = b                    \\\\
+                & lb \\leq x \\leq ub
+        \\end{array}\\end{split}
+
+    using one of the available QP solvers.
+
+    Parameters
+    ----------
+    R : numpy.array, scipy.sparse.csc_matrix or cvxopt.spmatrix
+        Symmetric matrix of the cost function (most solvers require it to be
+        definite).
+    s : numpy.array
+        Vector term of the cost function.
+    G : numpy.array, scipy.sparse.csc_matrix or cvxopt.spmatrix, optional
+        Linear inequality matrix.
+    h : numpy.array, optional
+        Linear inequality vector.
+    A : numpy.array, scipy.sparse.csc_matrix or cvxopt.spmatrix, optional
+        Linear equality matrix.
+    b : numpy.array, optional
+        Linear equality vector.
+    lb: numpy.array, scipy.sparse.csc_matrix or cvxopt.spmatrix, optional
+        Lower bound constraint vector.
+    ub: numpy.array, scipy.sparse.csc_matrix or cvxopt.spmatrix, optional
+        Upper bound constraint vector.
+    W : numpy.array, scipy.sparse.csc_matrix or cvxopt.spmatrix, optional
+        Definite symmetric weight matrix used to define the norm of the cost
+        function. The standard L2 norm (W = Identity) is used by default.
+    solver : string, optional
+        Name of the QP solver, to choose in ``qpsolvers.available_solvers``.
+    initvals : array, optional
+        Vector of initial `x` values used to warm-start the solver.
+    sym_proj : bool, optional
+        Set to `True` when the `R` matrix provided is not symmetric.
+    verbose : bool, optional
+        Set to `True` to print out extra information.
+
+    Returns
+    -------
+    x : array or None
+        Optimal solution if found, None otherwise.
+
+    Note
+    ----
+    Extra keyword arguments given to this function are forwarded to the
+    underlying solvers. For example, OSQP has a setting `eps_abs` which we can
+    provide by ``solve_ls(R, s, G, h, solver='osqp', eps_abs=1e-4)``.
+    """
+    if sym_proj:
+        R = 0.5 * (R + R.transpose())
+    WR = R if W is None else dot(W, R)
+    P = dot(R.transpose(), WR)
+    q = -dot(s.transpose(), WR)
+    return solve_qp(
+        P,
+        q,
+        G,
+        h,
+        A,
+        b,
+        lb,
+        ub,
+        solver=solver,
+        initvals=initvals,
+        sym_proj=False,
+        verbose=verbose,
+        **kwargs,
+    )
+
+
 __all__ = [
-    'available_solvers',
-    'cvxopt_solve_qp',
-    'cvxpy_solve_qp',
-    'dense_solvers',
-    'gurobi_solve_qp',
-    'mosek_solve_qp',
-    'qpoases_solve_qp',
-    'quadprog_solve_qp',
-    'solve_qp',
-    'solve_safer_qp',
-    'sparse_solvers',
+    "__version__",
+    "available_solvers",
+    "cvxopt_solve_qp",
+    "cvxpy_solve_qp",
+    "dense_solvers",
+    "gurobi_solve_qp",
+    "mosek_solve_qp",
+    "qpoases_solve_qp",
+    "quadprog_solve_qp",
+    "solve_ls",
+    "solve_qp",
+    "solve_safer_qp",
+    "sparse_solvers",
 ]
