@@ -18,41 +18,65 @@
 # You should have received a copy of the GNU Lesser General Public License
 # along with qpsolvers. If not, see <http://www.gnu.org/licenses/>.
 
+"""Solver interface for CVXOPT"""
+
+from typing import Optional
+
 from cvxopt import matrix, spmatrix
 from cvxopt.solvers import options, qp
 from numpy import array, ndarray
 
 
-options['show_progress'] = False  # disable cvxopt output
+options["show_progress"] = False  # disable cvxopt output
 
 
 def cvxopt_matrix(M):
-    if type(M) is ndarray:
+    """
+    Convert matrix to CVXOPT format.
+
+    Parameters
+    ----------
+    M : numpy.array
+        Matrix in NumPy format.
+
+    Returns
+    -------
+    N : cvxopt.matrix
+        Matrix in CVXOPT format.
+    """
+    if isinstance(M, ndarray):
         return matrix(M)
-    elif type(M) is spmatrix or type(M) is matrix:
+    if isinstance(M, (spmatrix, matrix)):
         return M
     coo = M.tocoo()
-    return spmatrix(
-        coo.data.tolist(), coo.row.tolist(), coo.col.tolist(), size=M.shape)
+    return spmatrix(coo.data.tolist(), coo.row.tolist(), coo.col.tolist(), size=M.shape)
 
 
-def cvxopt_set_verbosity(verbose):
-    options['show_progress'] = verbose
-
-
-def cvxopt_solve_qp(P, q, G=None, h=None, A=None, b=None, solver=None,
-                    initvals=None):
+def cvxopt_solve_qp(
+    P,
+    q,
+    G=None,
+    h=None,
+    A=None,
+    b=None,
+    solver: str = None,
+    initvals=None,
+    verbose: bool = False,
+) -> Optional[ndarray]:
     """
     Solve a Quadratic Program defined as:
 
-        minimize
-            (1/2) * x.T * P * x + q.T * x
+    .. math::
 
-        subject to
-            G * x <= h
-            A * x == b
+        \\begin{split}\\begin{array}{ll}
+        \\mbox{minimize} &
+            \\frac{1}{2} x^T P x + q^T x \\\\
+        \\mbox{subject to}
+            & G x \\leq h                \\\\
+            & A x = h
+        \\end{array}\\end{split}
 
-    using CVXOPT <http://cvxopt.org/>.
+    using `CVXOPT <http://cvxopt.org/>`_.
 
     Parameters
     ----------
@@ -72,10 +96,12 @@ def cvxopt_solve_qp(P, q, G=None, h=None, A=None, b=None, solver=None,
         Set to 'mosek' to run MOSEK rather than CVXOPT.
     initvals : numpy.array, optional
         Warm-start guess vector.
+    verbose : bool, optional
+        Set to `True` to print out extra information.
 
     Returns
     -------
-    x : array, shape=(n,)
+    x : array, shape=(n,), or None
         Solution to the QP, if found, otherwise ``None``.
 
     Note
@@ -83,15 +109,16 @@ def cvxopt_solve_qp(P, q, G=None, h=None, A=None, b=None, solver=None,
     CVXOPT only considers the lower entries of `P`, therefore it will use a
     wrong cost function if a non-symmetric matrix is provided.
     """
+    options["show_progress"] = verbose
     args = [cvxopt_matrix(P), cvxopt_matrix(q)]
-    kwargs = {'G': None, 'h': None, 'A': None, 'b': None}
+    kwargs = {"G": None, "h": None, "A": None, "b": None}
     if G is not None:
-        kwargs['G'] = cvxopt_matrix(G)
-        kwargs['h'] = cvxopt_matrix(h)
+        kwargs["G"] = cvxopt_matrix(G)
+        kwargs["h"] = cvxopt_matrix(h)
     if A is not None:
-        kwargs['A'] = cvxopt_matrix(A)
-        kwargs['b'] = cvxopt_matrix(b)
+        kwargs["A"] = cvxopt_matrix(A)
+        kwargs["b"] = cvxopt_matrix(b)
     sol = qp(*args, solver=solver, initvals=initvals, **kwargs)
-    if 'optimal' not in sol['status']:
+    if "optimal" not in sol["status"]:
         return None
-    return array(sol['x']).reshape((q.shape[0],))
+    return array(sol["x"]).reshape((q.shape[0],))
