@@ -18,28 +18,32 @@
 # You should have received a copy of the GNU Lesser General Public License
 # along with qpsolvers. If not, see <http://www.gnu.org/licenses/>.
 
-from numpy import hstack, vstack
+"""Solver interface for quadprog"""
+
+from typing import Optional
+from warnings import warn
+
+from numpy import hstack, ndarray, vstack
 from quadprog import solve_qp
 
 
-def quadprog_set_verbosity(verbose):
-    if verbose:
-        print("qpsolvers: quadprog solver does not provide a verbosity option")
-
-
-
-def quadprog_solve_qp(P, q, G=None, h=None, A=None, b=None, initvals=None):
+def quadprog_solve_qp(
+    P, q, G=None, h=None, A=None, b=None, initvals=None, verbose: bool = False
+) -> Optional[ndarray]:
     """
     Solve a Quadratic Program defined as:
 
-        minimize
-            (1/2) * x.T * P * x + q.T * x
+    .. math::
 
-        subject to
-            G * x <= h
-            A * x == b
+        \\begin{split}\\begin{array}{ll}
+        \\mbox{minimize} &
+            \\frac{1}{2} x^T P x + q^T x \\\\
+        \\mbox{subject to}
+            & G x \\leq h                \\\\
+            & A x = h
+        \\end{array}\\end{split}
 
-    using quadprog <https://pypi.python.org/pypi/quadprog/>.
+    using `quadprog <https://pypi.python.org/pypi/quadprog/>`_.
 
     Parameters
     ----------
@@ -57,6 +61,8 @@ def quadprog_solve_qp(P, q, G=None, h=None, A=None, b=None, initvals=None):
         Linear equality constraint vector.
     initvals : numpy.array, optional
         Warm-start guess vector (not used).
+    verbose : bool, optional
+        This argument has no effect, it is here for API conformance.
 
     Returns
     -------
@@ -69,13 +75,13 @@ def quadprog_solve_qp(P, q, G=None, h=None, A=None, b=None, initvals=None):
     will use a wrong cost function if a non-symmetric matrix is provided.
     """
     if initvals is not None:
-        print("quadprog: note that warm-start values ignored by wrapper")
+        warn("note that warm-start values ignored by this wrapper")
     qp_G = P
     qp_a = -q
     if A is not None:
         if G is None:
             qp_C = -A.T
-            qp_b = -b
+            qp_b = -b  # not None, checked in check_problem_constraints
         else:
             qp_C = -vstack([A, G]).T
             qp_b = -hstack([b, h])
@@ -87,7 +93,11 @@ def quadprog_solve_qp(P, q, G=None, h=None, A=None, b=None, initvals=None):
     try:
         return solve_qp(qp_G, qp_a, qp_C, qp_b, meq)[0]
     except ValueError as e:
-        if "matrix G is not positive definite" in e:
+        error = str(e)
+        if "matrix G is not positive definite" in error:
             # quadprog writes G the cost matrix that we write P in this package
-            raise ValueError("matrix P is not positive definite")
-        raise
+            raise ValueError("matrix P is not positive definite") from e
+        if "no solution" in error:
+            return None
+        warn("quadprog raised a ValueError: {}".format(e))
+        return None
